/**************************************************************************
   printoxx     print a set of graphic images

   Free software licensed under GNU General Public License v.2
   source URL:  kornelix.squarespace.com
***************************************************************************/

#include "zfuncs.h"

#define fversion "printoxx v.1.8.1  2009.01.28"                            //  version v.1.8.1
#define flicense "Free software - GNU General Public License v.2"
#define fsource "http://kornelix.squarespace.com/printoxx"

#define bilinear GDK_INTERP_BILINEAR                                       //  GDK shortcuts
#define nodither GDK_RGB_DITHER_NONE,0,0
#define colorspace GDK_COLORSPACE_RGB
#define pixbuf_scale gdk_pixbuf_scale_simple
#define angle GDK_PIXBUF_ROTATE_CLOCKWISE

GtkWidget   *mainWin, *drawWin, *mVbox;                                    //  main and drawing window
GdkPixbuf   *pxb_layout = 0;                                               //  print layout pixbuf
GdkPixbuf   *pxb_drawwin = 0;                                              //  drawing window pixbuf
GdkGC       *gdkgc = 0;                                                    //  graphics context
GError      **gerror = 0;                                                  //  pixbuf error code
int         wwD = 800, hhD = 600;                                          //  main window initial size
typedef     uchar *pixel;                                                  //  3 RGB values, 0-255 each

char        clfile[maxfcc] = "";                                           //  command line image file
char        imagefile[maxfcc] = "";                                        //  current image file
char        imagedirk[maxfcc] = "";                                        //  current image file directory

//  page layout data
int         layout_ww, layout_hh;                                          //  layout pixel dimensions
int         layout_pix = 400;                                              //  default image size in layout
int         layout_select = 0;                                             //  last selected image/text index
double      Rscale = 0.5;                                                  //  layout to drawing window
char        pendfile[maxfcc] = "";                                         //  pending image, add to layout
char        pendtext[1000] = "";                                           //  pending text, add to layout
char        textfont[60] = "sans 40";                                      //  layout text initial font
int         transp = 0;                                                    //  pending text: transparent
int         txcolor = 1;                                                   //  pending text: 1/2 = black/white

struct      playout {                                                      //  image and text members of layout
   int         type;                                                       //  1: file image, 2: text image
   int         txcolor;                                                    //  1: black on white, 2: reverse
   int         transp;                                                     //  text background is transparent
   char        *filespec;                                                  //  filespec or text
   GdkPixbuf   *pixbuf1;                                                   //  image pixbuf, full size
   GdkPixbuf   *pixbuf2;                                                   //  image pixbuf, rscale size
   int         ww, hh;                                                     //  full image dimensions
   int         xpos, ypos;                                                 //  position in layout
   double      rscale;                                                     //  scale factor, image to layout
};
playout     layout[200];                                                   //  layout for all images + text
int         maxNF = 200;                                                   //  max. images + text in layout
int         NPL = 0;                                                       //  current layout count

//  default printer setup
char     printer[60] = "default";                                          //  printer name
char     jobname[60] = "printoxx";                                         //  print job name
char     paper[40] = "A4  21.0 x 29.7 cm";                                 //  paper format and dimensions
char     orientation[20] = "portrait";                                     //  portrait / landscape
double   paperwidth, paperheight;                                          //  paper size, cm

//  known paper types and sizes                                            //  v.1.4
int         nptab = 7;
int         defptab = 1;                                                   //  default is ptab[1] = A4
const char  *ptab[7] = { "letter  21.6 x 27.9 cm" ,
                         "A4  21.0 x 29.7 cm" ,
                         "A5  14.8 x 21.0 cm" ,
                         "A6  10.5 x 14.8 cm" ,
                         "A7  7.4 x 10.5 cm" ,
                         "A8  5.2 x 7.4 cm" ,
                         "A9  3.7 x 5.2 cm"  };
   
//  GTK functions
int   gtkinitfunc(void *data);                                             //  GTK initz. function
void  mwpaint();                                                           //  window repaint - expose event
void  mwpaint2();                                                          //  window repaint - application call
void  mouse_event(GtkWidget *, GdkEventButton *, void *);                  //  mouse event function
int   drop_text(GtkWidget *, void *, int, int, void *);                    //  file drag/drop event function
void  menufunc(GtkWidget *, const char *menu);                             //  menu function, main window
void  destroy();                                                           //  main window destroy signal
int   KBpress(GtkWidget *, GdkEventKey *, void *);                         //  KB key press event
int   KBrelease(GtkWidget *, GdkEventKey *, void *);                       //  KB key release event

//  toolbar button functions
void  m_index();                                                           //  show thumbnail index window
void  m_image();                                                           //  add images to layout
void  m_text();                                                            //  add text to layout
void  m_rotate();                                                          //  rotate image 90 deg.
void  m_save();                                                            //  save layout to file
void  m_setup();                                                           //  print setup
void  m_print();                                                           //  print layout 
void  m_help();                                                            //  show user guide
void  m_quit();                                                            //  quit

void  init_layout();                                                       //  page layout from paper size
void  set_pendfile(char *file);                                            //  set pending file to be added
void  add_layout_file(int mpx, int mpy);                                   //  add pending file to layout
void  add_layout_text(int mpx, int mpy);                                   //  add pending text to layout
void  print_dialog();                                                      //  printer setup dialog
void  file_dialog();                                                       //  get image file dialog
void  text_dialog();                                                       //  get text dialog
void  load_imagedirk();                                                    //  load saved image directory
void  save_imagedirk();                                                    //  save for next startup

GdkPixbuf * make_layout1();                                                //  layout pixbuf for editing
GdkPixbuf * make_layout2();                                                //  layout pixbuf for printing
GdkPixbuf * load_pixbuf(const char *file);                                 //  load pixbuf from file, remove alpha chan.
GdkPixbuf * pixbuf_text(GtkWidget *, cchar *text, cchar *font, int color); //  create pixbuf containing text
void pixbuf_copy_alpha(GdkPixbuf *pxb1, int x1, int y1, int ww, int hh,    //  copy pixbuf1 into pixbuf2, treating
                       GdkPixbuf *pxb2, int x2, int y2, int color);        //    "color" in pixbuf1 as transparent


/**************************************************************************/

//  main program

int main(int argc, char *argv[])
{
   GtkWidget   *mtbar;
   int         ii;
   char        lang[4] = "", string[] = "STRING";
   
   GtkTargetEntry file_drop_target = { string, 0, 0 };

   gtk_init(&argc,&argv);                                                  //  GTK command line options

   initz_appfiles("printoxx",null);                                        //  get app directories
   zlockInit();

   for (ii = 1; ii < argc; ii++)                                           //  command line options
   {
      if (strEqu(argv[ii],"-l") && argc > ii+1)                            //  -l language code
            strncpy0(lang,argv[++ii],3);
      else if (strEqu(argv[ii],"-v")) {                                    //  -v print version and exit
            printf(fversion "\n"); return 0; }
      else if (strEqu(argv[ii],"-i") && argc > ii+1)                       //  -i imageDirectory
            strcpy(imagedirk,argv[++ii]);
      else if (strEqu(argv[ii],"-f") && argc > ii+1)                       //  -f imageFile
            strcpy(clfile,argv[++ii]);
      else strcpy(clfile,argv[ii]);                                        //  assume imageFile
   }
   
   ZTXinit(lang);                                                          //  setup translations

   mainWin = gtk_window_new(GTK_WINDOW_TOPLEVEL);                          //  create main window
   gtk_window_set_title(GTK_WINDOW(mainWin),fversion);
   gtk_window_set_position(GTK_WINDOW(mainWin),GTK_WIN_POS_CENTER);
   gtk_window_set_default_size(GTK_WINDOW(mainWin),wwD,hhD);

   mVbox = gtk_vbox_new(0,0);                                              //  add vert. packing box
   gtk_container_add(GTK_CONTAINER(mainWin),mVbox);

   mtbar = create_toolbar(mVbox,24);                                       //  add toolbar and buttons  
   add_toolbar_button(mtbar,ZTX("index"),ZTX("thumbnail index"),"index.png",menufunc);
   add_toolbar_button(mtbar,ZTX("+image"),ZTX("add image"),"image.png",menufunc);
   add_toolbar_button(mtbar,ZTX("+text"),ZTX("add text"),"text.png",menufunc);
   add_toolbar_button(mtbar,ZTX("rotate"),ZTX("rotate image"),"rotate.png",menufunc);
   add_toolbar_button(mtbar,ZTX("toolbar::save"),ZTX("save layout to file"),"gtk-save",menufunc);
   add_toolbar_button(mtbar,ZTX("setup"),ZTX("print setup"),"print.png",menufunc);
   add_toolbar_button(mtbar,ZTX("print"),ZTX("print page"),"print.png",menufunc);
   add_toolbar_button(mtbar,ZTX("quit"),ZTX("quit printoxx"),"gtk-quit",menufunc);
   add_toolbar_button(mtbar,ZTX("help"),ZTX("help"),"gtk-help",menufunc);

   drawWin = gtk_drawing_area_new();                                       //  drawing window
   gtk_container_add(GTK_CONTAINER(mVbox),drawWin);                        //  add to main window

   G_SIGNAL(mainWin,"destroy",destroy,0)                                   //  connect events to main window
   G_SIGNAL(drawWin,"expose-event",mwpaint,0)

   G_SIGNAL(mainWin,"key-press-event",KBpress,0)                           //  connect KB events
   G_SIGNAL(mainWin,"key-release-event",KBrelease,0)

   gtk_widget_add_events(drawWin,GDK_BUTTON_PRESS_MASK);                   //  connect mouse events
   gtk_widget_add_events(drawWin,GDK_BUTTON_RELEASE_MASK);
   gtk_widget_add_events(drawWin,GDK_BUTTON_MOTION_MASK);
   gtk_widget_add_events(drawWin,GDK_POINTER_MOTION_MASK);
   G_SIGNAL(drawWin,"button-press-event",mouse_event,0)
   G_SIGNAL(drawWin,"button-release-event",mouse_event,0)
   G_SIGNAL(drawWin,"motion-notify-event",mouse_event,0)

   gtk_drag_dest_set(drawWin,GTK_DEST_DEFAULT_ALL,&file_drop_target,1,     //  connect drag/drop event   v.1.8
                                                   GDK_ACTION_COPY);
   G_SIGNAL(drawWin,"drag-data-received",drop_text,0)
   
   gtk_widget_show_all(mainWin);                                           //  show all widgets
   gdkgc = gdk_gc_new(drawWin->window);                                    //  initz. graphics context
   
   gtk_init_add((GtkFunction) gtkinitfunc,0);                              //  initz. function
   gtk_main();                                                             //  process window events

   return 0;
}


//  initial function called from gtk_main() at startup

int gtkinitfunc(void * data)
{
   char        *ppv, clfile2[maxfcc], file[200];
   const char  *pp;
   int         ii, ignore;
   FILE        *fid;

   snprintf(file,200,"%s/print-setup",get_zuserdir());                     //  get prior printer setups
   fid = fopen(file,"r");
   if (fid) {
      ignore = fscanf(fid,"printer %s ",printer);
      ignore = fscanf(fid,"jobname %s ",jobname);
      ignore = fscanf(fid,"paper   %[ -z] ",paper);
      ignore = fscanf(fid,"orientation %s ",orientation);
      fclose(fid);
   }
   strTrim(paper);
   
   for (ii = 0; ii < nptab; ii++)                                          //  match paper type with table
      if (strEqu(paper,ptab[ii])) break;
   if (ii == nptab) ii = defptab;                                          //  not found, use default
   strcpy(paper,ptab[ii]);
   pp = strField(paper,' ',2);
   paperwidth = atoi(pp);
   pp = strField(paper,' ',4);
   paperheight = atoi(pp);
   
   if (strNeq(orientation,"portrait") &&                                   //  check/set orientation
       strNeq(orientation,"landscape"))
         strcpy(orientation,"portrait");
   
   init_layout();                                                          //  set up page layout pixbuf
   
   if (! *imagedirk) load_imagedirk();                                     //  recover image directory   v.1.3

   if (*clfile) {                                                          //  command line image file
      if (*clfile != '/') {
         strcpy(clfile2,imagedirk);                                        //  clfile relative to imagedirk
         strcat(clfile2,"/");
         strcat(clfile2,clfile);
      }
      else { 
         strcpy(clfile2,clfile);                                           //  clfile is full pathspec
         strcpy(imagedirk,clfile);
         ppv = strrchr(imagedirk,'/');
         if (ppv) *ppv = 0;
      }

      set_pendfile(clfile2);                                               //  check if valid image file
      if (*pendfile) add_layout_file(100,100);                             //  add to layout, upper left corner
   }

   return 0;
}


//  paint window when created, exposed, resized

void mwpaint()
{
   GdkPixbuf   *pxbM;
   GdkDrawable *Lwin = drawWin->window;
   int         ww, hh;
   double      scalew, scaleh;
   
   if (pxb_layout) g_object_unref(pxb_layout);
   pxb_layout = make_layout1();

   wwD = drawWin->allocation.width;                                        //  current drawing window size
   hhD = drawWin->allocation.height;

   scalew = 1.0 * wwD / layout_ww;                                         //  scale layout to window
   scaleh = 1.0 * hhD / layout_hh;
   if (scalew < scaleh) Rscale = scalew;
   else Rscale = scaleh;
   ww = int(Rscale * layout_ww + 0.5);
   hh = int(Rscale * layout_hh + 0.5);
   pxbM = pixbuf_scale(pxb_layout,ww,hh,bilinear);
   if (! pxbM) zappcrash("cannot create pixbuf");

   gdk_draw_pixbuf(Lwin,0,pxbM,0,0,0,0,-1,-1,nodither);
   g_object_unref(pxbM);

   return;
}


//  invalidate window - cause repaint for application changes

void mwpaint2()
{
   gdk_window_invalidate_rect(drawWin->window,0,1);                        //  invalidate whole window   v.1.8
   return;
}


//  mouse event function - capture buttons and drag movements

void mouse_event(GtkWidget *, GdkEventButton *event, void *)
{
   static int     bdtime = 0, butime = 0;
   static int     Lmouse = 0, Rmouse = 0;
   static int     Fclick = 0, Fselect = 0, Fcorner = 0;
   static int     elapsed, mpx0 = 0, mpy0 = 0;
   static int     ii, kk, ww, hh;
   int            mpx, mpy, dx = 0, dy = 0;
   int            x1, y1, x2, y2;
   double         rscale;
   playout        templayout;

   mpx = int(event->x);                                                    //  mouse position in window
   mpy = int(event->y);
   mpx = int(mpx / Rscale + 0.5);                                          //  in print layout
   mpy = int(mpy / Rscale + 0.5);

   if (event->type == GDK_BUTTON_PRESS)
   {
      Lmouse = Rmouse = 0;
      if (event->button == 1) Lmouse++;                                    //  left or right mouse button
      if (event->button == 3) Rmouse++;
      bdtime = event->time;
      Fselect = Fcorner = 0;

      for (ii = NPL-1; ii >= 0; ii--)                                      //  search from top image down
      {
         rscale = layout[ii].rscale;
         ww = int(rscale * layout[ii].ww);                                 //  is cursor inside an image?
         if (mpx < layout[ii].xpos) continue;
         if (mpx > layout[ii].xpos + ww) continue;
         hh = int(rscale * layout[ii].hh);
         if (mpy < layout[ii].ypos) continue;
         if (mpy > layout[ii].ypos + hh) continue;
         break;
      }
      if (ii >= 0) {                                                       //  yes, iith image/text selected
         layout_select = ii;                                               //  remember last selected object
         if (layout[ii].type == 1)                                         //  if image, set curr. image file
            strcpy(imagefile,layout[ii].filespec);                         //  (potential thumbnail index anchor)
         Fselect = 1;
      }
      
      if (Fselect) 
      {
         mpx0 = mpx;                                                       //  set new drag origin
         mpy0 = mpy;
         
         if (layout[ii].type == 1) {                                       //  mouse in image corner?
            if (mpx < layout[ii].xpos + 0.2 * ww  &&
                mpy < layout[ii].ypos + 0.2 * hh) Fcorner = 1;             //  upper left
            if (mpx > layout[ii].xpos + 0.8 * ww  &&
                mpy < layout[ii].ypos + 0.2 * hh) Fcorner = 2;             //  upper right
            if (mpx > layout[ii].xpos + 0.8 * ww  &&
                mpy > layout[ii].ypos + 0.8 * hh) Fcorner = 3;             //  lower right
            if (mpx < layout[ii].xpos + 0.2 * ww  &&
                mpy > layout[ii].ypos + 0.8 * hh) Fcorner = 4;             //  lower left
         }

         if (layout[ii].type == 2) {                                       //  mouse in text corner?
            if (ww > hh) {                                                 //  (broader range for narrow edge)
               if (mpx < layout[ii].xpos + 0.2 * ww  &&
                   mpy < layout[ii].ypos + 0.5 * hh) Fcorner = 1;          //  upper left
               if (mpx > layout[ii].xpos + 0.8 * ww  &&
                   mpy < layout[ii].ypos + 0.5 * hh) Fcorner = 2;          //  upper right
               if (mpx > layout[ii].xpos + 0.8 * ww  &&
                   mpy > layout[ii].ypos + 0.5 * hh) Fcorner = 3;          //  lower right
               if (mpx < layout[ii].xpos + 0.2 * ww  &&
                   mpy > layout[ii].ypos + 0.5 * hh) Fcorner = 4;          //  lower left
            }

            else  {
               if (mpx < layout[ii].xpos + 0.5 * ww  &&
                   mpy < layout[ii].ypos + 0.2 * hh) Fcorner = 1;          //  upper left
               if (mpx > layout[ii].xpos + 0.5 * ww  &&
                   mpy < layout[ii].ypos + 0.2 * hh) Fcorner = 2;          //  upper right
               if (mpx > layout[ii].xpos + 0.5 * ww  &&
                   mpy > layout[ii].ypos + 0.8 * hh) Fcorner = 3;          //  lower right
               if (mpx < layout[ii].xpos + 0.5 * ww  &&
                   mpy > layout[ii].ypos + 0.8 * hh) Fcorner = 4;          //  lower left
            }
         }
      }
      
      return;
   }
   
   if (event->type == GDK_MOTION_NOTIFY)                                   //  capture mouse movement
   {
      if (! Lmouse) return;                                                //  only if left-mouse

      if (Fselect) {
         dx = mpx - mpx0;                                                  //  capture drag motion
         dy = mpy - mpy0;
         mpx0 = mpx;                                                       //  set new drag origin
         mpy0 = mpy;
      }
      
      if (Fselect && ! Fcorner) {                                          //  image move: add mouse drag
         layout[ii].xpos += dx;                                            //    to image position
         layout[ii].ypos += dy;
      }
      
      if (Fcorner)                                                         //  image resize: add mouse drag
      {                                                                    //    to image size
         rscale = layout[ii].rscale;
         ww = layout[ii].ww;
         hh = layout[ii].hh;
         x1 = layout[ii].xpos;
         y1 = layout[ii].ypos;
         x2 = x1 + int(rscale * ww + 0.5);
         y2 = y1 + int(rscale * hh + 0.5);

         if (Fcorner == 1) {                                               //  drag upper left corner
            if (ww > hh) rscale = (rscale * ww - dx) / ww;
            else  rscale = (rscale * hh - dy) / hh;
            x1 = x2 - int(rscale * ww + 0.5);                              //  leave opposite corner fixed
            y1 = y2 - int(rscale * hh + 0.5);
         }

         if (Fcorner == 2) {                                               //  upper right
            if (ww > hh) rscale = (rscale * ww + dx) / ww;
            else  rscale = (rscale * hh - dy) / hh;
            x2 = x1 + int(rscale * ww + 0.5);
            y1 = y2 - int(rscale * hh + 0.5);
         }

         if (Fcorner == 3) {                                               //  lower right
            if (ww > hh) rscale = (rscale * ww + dx) / ww;
            else  rscale = (rscale * hh + dy) / hh;
            x2 = x1 + int(rscale * ww + 0.5);
            y2 = y1 + int(rscale * hh + 0.5);
         }

         if (Fcorner == 4) {                                               //  lower left
            if (ww > hh) rscale = (rscale * ww - dx) / ww;
            else  rscale = (rscale * hh + dy) / hh;
            x1 = x2 - int(rscale * ww + 0.5);
            y2 = y1 + int(rscale * hh + 0.5);
         }
         
         if (rscale < 0.05 || rscale > 4.0) return;                        //  limit extremes
         
         if (ww > hh) rscale = 1.0 * (x2 - x1) / ww;                       //  make precise
         else  rscale = 1.0 * (y2 - y1) / hh;

         if (layout[ii].pixbuf2) g_object_unref(layout[ii].pixbuf2);       //  discard old scaled pixbuf
         layout[ii].pixbuf2 = 0;

         layout[ii].rscale = rscale;                                       //  will recreate from new data
         layout[ii].xpos = x1;
         layout[ii].ypos = y1;
      }
   }
   
   if (event->type == GDK_BUTTON_RELEASE)
   {
      Fclick = 0;
      butime = event->time;
      elapsed = butime - bdtime;                                           //  button down time, milliseconds
      if (elapsed < 500) Fclick = 1;                                       //  mouse clicked, no drag

      if (Fselect && Fclick && Lmouse) {                                   //  image was left-clicked
         templayout = layout[ii];
         for (kk = ii; kk < NPL-1; kk++)                                   //  make topmost (will occlude others)
         {
            if (layout[kk+1].type > layout[ii].type) break;                //  all text above all images   v.1.6
            layout[kk] = layout[kk+1];
         }
         layout[kk] = templayout;
         ii = layout_select = kk;                                          //  new current image index     v.1.7
      }

      if (Fselect && Fclick && Rmouse) {                                   //  image was right-clicked
         zfree(layout[ii].filespec);
         g_object_unref(layout[ii].pixbuf1);                               //  delete image from layout
         g_object_unref(layout[ii].pixbuf2);
         for (kk = ii; kk < NPL-1; kk++)                                   //  cover the hole
            layout[kk] = layout[kk+1];
         NPL--;
      }

      if (! Fselect && ! Fcorner && Fclick) {                              //  empty space was clicked
         if (*pendfile) add_layout_file(mpx,mpy);                          //  add pending file image to layout
         else if (*pendtext) add_layout_text(mpx,mpy);                     //  add pending text image to layout
      }
      
      Fselect = Fcorner = Fclick = 0;                                      //  stop drag/resize
      Lmouse = Rmouse = 0;
   }

   mwpaint2();                                                             //  repaint window
   return;
}


//  text drag and drop event - get filespec or text to add to layout       v.1.8

int drop_text(GtkWidget *, void *, int mpx, int mpy, void *sdata)
{
   char     *text, *file;
   int      cc;
   
   text = (char *) ((GtkSelectionData *) sdata)->data;

   mpx = int(mpx / Rscale + 0.5);                                          //  in print layout
   mpy = int(mpy / Rscale + 0.5);

   if (strstr(text,"file://")) 
   {
      file = strdupz(text+7);                                              //  get rid of junk added by GTK
      cc = strlen(file);
      while (file[cc-1] < ' ') cc--;
      file[cc] = 0;
      clean_percent_escapes(file);                                         //  blanks may be  %20, etc. 

      set_pendfile(file);                                                  //  check if valid image file
      *pendtext = 0;
      if (*pendfile) add_layout_file(mpx,mpy);                             //  add to layout at position
      zfree(file);
   }
   
   else 
   {      
      strncpy0(pendtext,text,999);                                         //  save for layout insertion
      *pendfile = 0;
      add_layout_text(mpx,mpy);
   }

   mwpaint2();
   return 1;
}


//  prevent propagation of key-press events to toolbar buttons     v.1.7

int KBpress(GtkWidget *win, GdkEventKey *event, void *)
{
   return 1;
}


//  keyboard event - arrow keys >> 1-pixel movement of current image
//  GDK key symbols: /usr/include/gtk-2.0/gdk/gdkkeysyms.h

int KBrelease(GtkWidget *win, GdkEventKey *event, void *)                  //  process KB arrow keys   v.1.7
{
   int         ii, KBkey;

   ii = layout_select;
   if (ii < 0 || ii >= NPL) return 1;

   KBkey = event->keyval;
   if (KBkey == GDK_Left) --layout[ii].xpos;                               //  move image in 1-pixel steps
   if (KBkey == GDK_Right) ++layout[ii].xpos;
   if (KBkey == GDK_Up) --layout[ii].ypos;
   if (KBkey == GDK_Down) ++layout[ii].ypos;
   
   mwpaint2();
   return 1;
}


//  process main window toolbar events

void menufunc(GtkWidget *, const char *menu)
{
   if (strEqu(menu,ZTX("index"))) { m_index(); return; }
   if (strEqu(menu,ZTX("+image"))) { m_image(); return; }
   if (strEqu(menu,ZTX("+text"))) { m_text(); return; }
   if (strEqu(menu,ZTX("rotate"))) { m_rotate(); return; }
   if (strEqu(menu,ZTX("save"))) { m_save(); return; }
   if (strEqu(menu,ZTX("setup"))) { m_setup(); return; }
   if (strEqu(menu,ZTX("print"))) { m_print(); return; }
   if (strEqu(menu,ZTX("quit"))) { m_quit(); return; }
   if (strEqu(menu,ZTX("help"))) { m_help(); return; }
   return;
}


//  main window destroy signal

void destroy()
{
   printf("main window destroyed \n");
   save_imagedirk();
   gtk_main_quit();   
   return;
}


//  display a thumbnail index of images in a separate window

void m_index()                                                             //  v.1.3.2
{
   static char    pimagedirk[maxfcc] = "";
   
   if (strNeq(imagedirk,pimagedirk)) {
      image_xthumbs(imagedirk,"init",0,set_pendfile);
      strcpy(pimagedirk,imagedirk);
   }

   if (*imagefile) image_xthumbs(imagefile,"paint1",0,set_pendfile);
   else  image_xthumbs(0,"paint1");

   return;
}


//  choose image files and add to print layout

void m_image()
{
   file_dialog();
   return;
}


//  add text to print layout

void m_text()
{
   text_dialog();
   return;
}


//  rotate last-selected image 90 degrees

void m_rotate()
{
   GdkPixbuf   *pixbuf; 
   int         ii, ww, hh;

   ii = layout_select;                                                     //  use last selected object   v.1.6
   if (ii < 0 || ii >= NPL) return;
   pixbuf = gdk_pixbuf_rotate_simple(layout[ii].pixbuf1,angle);
   if (! pixbuf) zappcrash("pixbuf rotate failed");
   g_object_unref(layout[ii].pixbuf1);
   layout[ii].pixbuf1 = pixbuf;
   g_object_unref(layout[ii].pixbuf2);
   layout[ii].pixbuf2 = 0;
   ww = layout[ii].ww;
   hh = layout[ii].hh;
   layout[ii].ww = hh;
   layout[ii].hh = ww;
   
   mwpaint2();
   return;
}


//  save page layout to a file

void m_save()
{
   GdkPixbuf   *layout2;
   char        printfile[200], command[maxfcc+100];
   char        file1[maxfcc], *file2;
   int         ignore;
   
   layout2 = make_layout2();                                               //  make 200% layout
   snprintf(printfile,199,"%s/printfile.jpg",get_zuserdir());              //  jpeg print file
   gdk_pixbuf_save(layout2,printfile,"jpeg",gerror,null);                  //  layout >> print file
   g_object_unref(layout2);
   
   snprintf(file1,maxfcc,"%s/printoxx.jpg",imagedirk);
   
   file2 = zgetfile(ZTX("save file"),file1,"save");                        //  query user for file
   if (! file2) return;

   snprintf(command,maxfcc+100,"cp -f \"%s\" \"%s\" ",printfile,file2);
   ignore = system(command);
   return;
}


//  printer setup (printer, paper size, etc.)

void m_setup()
{
   print_dialog();
   init_layout();
   return;
}


//  print the completed print layout

void m_print()
{
   GdkPixbuf   *layout2;
   char        printfile[200], command[200], papertype[100];
   const char  *pformat;
   int         ignore;
   
   layout2 = make_layout2();                                               //  make 200% layout
   snprintf(printfile,199,"%s/printfile.jpg",get_zuserdir());              //  jpeg print file
   gdk_pixbuf_save(layout2,printfile,"jpeg",gerror,null);                  //  layout >> print file
   g_object_unref(layout2);
   
   strcpy(command,"lp ");
   if (strNeq(printer,"default")) 
      strncatv(command,199,"-d ",printer," ",0);                           //  v.1.5
   
   pformat = strField(paper,' ',1);
   snprintf(papertype,99," -o PageSize=%s -o Media=%s ",pformat,pformat);
   strcat(command,papertype);
   strcat(command," -o scaling=98 ");                                      //  fitplot fails   v.1.4.1
   strcat(command,printfile);
   ignore = system(command);
   printf("printoxx: %s \n",command);
   return;
}


//  thread functions to display help file

void m_help()
{
   showz_helpfile();
   return;
}


//  quit - exit program

void m_quit()
{
   save_imagedirk();
   gtk_main_quit();
   return;
}


//  set up page layout pixbuf based on printer paper dimensions

void init_layout()
{
   if (strEqu(orientation,"portrait")) {
      layout_ww = 1000;
      layout_hh = int(paperheight * 1000 / paperwidth);
   }
   else {
      layout_ww = 1500;
      layout_hh = int(paperwidth * 1500 / paperheight);
   }
      
   pxb_layout = gdk_pixbuf_new(colorspace,0,8,layout_ww,layout_hh);        //  pixbuf for page layout

   wwD = int(Rscale * layout_ww);                                          //  matching drawing window size
   hhD = int(Rscale * layout_hh + 60);
   gtk_window_resize(GTK_WINDOW(mainWin),wwD,hhD);

   return;
}


//  verify chosen file and make pending add to layout
 
void set_pendfile(char *file)
{
   GdkPixbuf      *pixbuf;
   char           *pp;

   if (NPL == maxNF-1) {
      zmessageACK(ZTX("exceed %d files"),maxNF);
      return;
   }
   
   pixbuf = load_pixbuf(file);                                             //  test file is legit. image file
   if (! pixbuf) {
      zmessageACK(ZTX("not a compatible image file \n %s"),file);
      *pendfile = 0;                                                       //  v.1.2
      return;
   }
   g_object_unref(pixbuf);

   strcpy(pendfile,file);                                                  //  save filespec
   *pendtext = 0;

   strcpy(imagefile,file);                                                 //  set curr. image file
   strcpy(imagedirk,file);                                                 //  and image directory
   pp = strrchr(imagedirk,'/');
   if (pp) *pp = 0;
   
   gtk_window_present(GTK_WINDOW(mainWin));                                //  bring to foreground
   return;
}


//  add a new image file to the layout at designated position

void add_layout_file(int mpx, int mpy)
{
   GdkPixbuf   *pixbuf;
   int         ii, kk;

   pixbuf = load_pixbuf(pendfile);
   if (! pixbuf) zappcrash("cannot create pixbuf");
   for (ii = 0; ii < NPL; ii++)                                            //  insert at end of images    v.1.6
      if (layout[ii].type != 1) break;                                     //  (topmost file image)
   for (kk = NPL; kk > ii; kk--)
      layout[kk] = layout[kk-1];
   NPL++;
   layout_select = ii;
   layout[ii].type = 1;
   layout[ii].filespec = strdupz(pendfile);
   layout[ii].pixbuf1 = pixbuf;
   layout[ii].pixbuf2 = 0;
   layout[ii].ww = gdk_pixbuf_get_width(layout[ii].pixbuf1);
   layout[ii].hh = gdk_pixbuf_get_height(layout[ii].pixbuf1);
   layout[ii].xpos = mpx;
   layout[ii].ypos = mpy;
   layout[ii].rscale = 1.0 * layout_pix / layout[ii].ww;
   
   return;
}


//  add a new text image to the layout at designated position

void add_layout_text(int mpx, int mpy)
{
   GdkPixbuf   *pixbuf;
   int         ii;

   pixbuf = pixbuf_text(drawWin,pendtext,textfont,txcolor);
   if (! pixbuf) zappcrash("cannot create text pixbuf");
   ii = NPL++;                                                             //  topmost text image
   layout_select = ii;
   layout[ii].type = 2;
   layout[ii].txcolor = txcolor;
   layout[ii].transp = transp;
   layout[ii].filespec = strdupz(pendtext);
   layout[ii].pixbuf1 = pixbuf;
   layout[ii].pixbuf2 = 0;
   layout[ii].ww = gdk_pixbuf_get_width(layout[ii].pixbuf1);
   layout[ii].hh = gdk_pixbuf_get_height(layout[ii].pixbuf1);
   layout[ii].xpos = mpx;
   layout[ii].ypos = mpy;
   layout[ii].rscale = 0.5;

   return;
}


//  print setup dialog

void print_dialog()
{
   void  print_dialog_stuff(zdialog *zd);
   void  print_dialog_fetch(zdialog *zd);

   int         zstat, contx = 0;
   const char  *pp1, *pp2;

   zdialog *zd = zdialog_new(ZTX("print setup"),mainWin,ZTX("done"),ZTX("cancel"),null);
   zdialog_add_widget(zd,"vbox","vb1","dialog",0,"space=5");

   zdialog_add_widget(zd,"hbox","hb1","vb1",0,"space=5");
   zdialog_add_widget(zd,"vbox","vb11","hb1",0,"homog");                   //  job name    [__________]
   zdialog_add_widget(zd,"vbox","vb12","hb1",0,"homog");                   //  printer     [__________][v]
   zdialog_add_widget(zd,"label","labjob","vb11",ZTX("job name"));         //  paper type  [__________][v]
   zdialog_add_widget(zd,"label","labpri","vb11",ZTX("printer")); 
   zdialog_add_widget(zd,"label","labpap","vb11",ZTX("paper type"));
   zdialog_add_widget(zd,"entry","entjob","vb12","printoxx","scc=20");
   zdialog_add_widget(zd,"combo","entpri","vb12",0,"scc=20");
   zdialog_add_widget(zd,"combo","entpap","vb12",0,"scc=20");
   zdialog_add_widget(zd,"hbox","hb3","vb1",0,"space=10");                 //  (o) portrait  (o) landscape
   zdialog_add_widget(zd,"radio","port","hb3",ZTX("portrait"));
   zdialog_add_widget(zd,"radio","land","hb3",ZTX("landscape"));

   zdialog_cb_app(zd,"entpri","default");                                  //  list avail. printers  v.1.5
   while (true) {   
      pp1 = command_output(contx,"lpstat -a");
      if (! pp1) break;
      if (*pp1 <= ' ') continue;
      pp2 = strField(pp1,' ',1);
      zdialog_cb_app(zd,"entpri",pp2);
   }
   
   for (int ii = 0; ii < nptab; ii++)                                      //  stuff combobox with paper types
      zdialog_cb_app(zd,"entpap",ptab[ii]);
   
   print_dialog_stuff(zd);                                                 //  old data >> dialog
   zstat = zdialog_run(zd);                                                //  run dialog, blocking
   if (zstat == 1) print_dialog_fetch(zd);                                 //  dialog >> new data >> file
   zdialog_free(zd);
   return;
}


//  prior setup data >> dialog widgets

void print_dialog_stuff(zdialog *zd)                                       //  v.1.4
{
   zdialog_stuff(zd,"entpri",printer);                                     //  load dialog widgets with
   zdialog_stuff(zd,"entjob",jobname);                                     //    prior setup data
   zdialog_stuff(zd,"entpap",paper);
   zdialog_stuff(zd,"port",1);
   if (strEqu(orientation,"landscape")) zdialog_stuff(zd,"land",1);
   return;
}


//  dialog widgets >> new setup data >> file

void print_dialog_fetch(zdialog *zd)                                       //  v.1.4
{
   char        file[200];
   const char  *pp;
   int         ii, port;
   FILE        *fid;

   zdialog_fetch(zd,"entpri",printer,60);                                  //  retrieve data from widgets
   zdialog_fetch(zd,"entjob",jobname,60);
   zdialog_fetch(zd,"entpap",paper,40);
   zdialog_fetch(zd,"port",port);
   if (port) strcpy(orientation,"portrait");
   else strcpy(orientation,"landscape");

   snprintf(file,200,"%s/print-setup",get_zuserdir());
   fid = fopen(file,"w");                                                  //  write to file
   if (! fid) {
      zmessageACK(ZTX("cannot write print setup file: %s"),file);
      return;
   }

   fprintf(fid,"printer       %s \n",printer);
   fprintf(fid,"jobname       %s \n",jobname);
   fprintf(fid,"paper         %s \n",paper);
   fprintf(fid,"orientation   %s \n",orientation);
   fclose(fid);
   
   for (ii = 0; ii < nptab; ii++)                                          //  match paper type with table
      if (strEqu(paper,ptab[ii])) break;
   if (ii == nptab) ii = defptab;                                          //  not found, use default
   pp = strField(paper,' ',2);
   paperwidth = atoi(pp);
   pp = strField(paper,' ',4);
   paperheight = atoi(pp);
   return;
}


//  file-chooser dialog box and response function
//  chosen file is saved for drawing window deposit via mouse-click

void file_dialog()
{
   int file_event(GtkDialog *dwin, int arg, void *data);
   void file_signal(GtkFileChooser *file_widget, void *);
   void file_preview(GtkFileChooser *, GtkWidget *pvwidget);

   GtkWidget   *file_dialogbox, *file_widget;
   GtkWidget   *pvwidget = gtk_image_new();

   file_dialogbox = gtk_dialog_new_with_buttons(ZTX("choose files"),
                        GTK_WINDOW(mainWin),(GtkDialogFlags) 0,ZTX("done"),2,null);
   gtk_window_set_default_size(GTK_WINDOW(file_dialogbox),700,500);
   G_SIGNAL(file_dialogbox,"response",file_event,0)

   file_widget = gtk_file_chooser_widget_new(GTK_FILE_CHOOSER_ACTION_OPEN);
   gtk_container_add(GTK_CONTAINER(GTK_DIALOG(file_dialogbox)->vbox),file_widget);
   gtk_file_chooser_set_current_folder(GTK_FILE_CHOOSER(file_widget),imagedirk);
   gtk_file_chooser_set_show_hidden(GTK_FILE_CHOOSER(file_widget),0);
   G_SIGNAL(file_widget,"file_activated",file_signal,0)

   gtk_file_chooser_set_preview_widget(GTK_FILE_CHOOSER(file_widget),pvwidget);
   G_SIGNAL(file_widget,"update-preview",file_preview,pvwidget);

   gtk_widget_show_all(file_dialogbox);
   return;
}

int file_event(GtkDialog *dwin, int arg, void *data)                       //  done or cancel
{
   gtk_widget_destroy(GTK_WIDGET(dwin));
   return 0;
}

void file_signal(GtkFileChooser *file_widget, void *)                      //  get a chosen file
{
   char   *file;
   file = gtk_file_chooser_get_filename(file_widget);
   set_pendfile(file);                                                     //  save for layout insertion
   g_free(file);
   gtk_file_chooser_unselect_all(file_widget);
   return;
}

void file_preview(GtkFileChooser *file_widget, GtkWidget *pvwidget)        //  get file preview thumbnail
{
   GdkPixbuf   *thumbnail;
   char        *filename;

   filename = gtk_file_chooser_get_preview_filename(file_widget);
   thumbnail = image_thumbnail(filename,128);                              //  get 128x128 thumbnail   v.1.7
   g_free(filename);                                                       //  (cached or created)

   if (thumbnail) {
      gtk_image_set_from_pixbuf(GTK_IMAGE(pvwidget),thumbnail);
      gtk_file_chooser_set_preview_widget_active(file_widget,1);
      g_object_unref(thumbnail);
   }
   else
      gtk_file_chooser_set_preview_widget_active(file_widget,0);
   return;
}


//  get text dialog box and response function
//  entered text is saved for drawing window deposit via mouse-click

void text_dialog()
{
   int text_dialog_event(zdialog *zd, const char *event);
   int text_dialog_compl(zdialog *zd, int zstat);
   
   const char *helptext = ZTX(" enter text, select font, \n"
                              " click empty space on layout ");

   zdialog        *zd;

   zd = zdialog_new(ZTX("enter text"),mainWin,ZTX("cancel"),null);
   zdialog_add_widget(zd,"label","lab1","dialog",helptext,"space=10");     //  text [_______________]
   zdialog_add_widget(zd,"hbox","hb1","dialog",0,"space=5");
   zdialog_add_widget(zd,"label","lab2","hb1",ZTX("text"),"space=3");
   zdialog_add_widget(zd,"frame","fr1","hb1",0,"expand");
   zdialog_add_widget(zd,"edit","text","fr1",pendtext,"space=5");

   zdialog_add_widget(zd,"hbox","hb2","dialog");                           //  (o) black   (o) white
   zdialog_add_widget(zd,"radio","black","hb2",ZTX("black"),"space=5");
   zdialog_add_widget(zd,"radio","white","hb2",ZTX("white"),"space=10");

   zdialog_add_widget(zd,"hbox","hb3","dialog");                           //  transparent [_]
   zdialog_add_widget(zd,"label","lab3","hb3",ZTX("transparent"),"space=8");
   zdialog_add_widget(zd,"check","transp","hb3",0);

   zdialog_add_widget(zd,"hbox","hb4","dialog",0,"space=10");              //  [erase text]  [select font] 
   zdialog_add_widget(zd,"button","erase","hb4",ZTX("erase text"));
   zdialog_add_widget(zd,"button","font","hb4",ZTX("select font"));
   
   if (transp) zdialog_stuff(zd,"transp",1);
   if (txcolor == 1) zdialog_stuff(zd,"black",1);
   else  zdialog_stuff(zd,"white",1);

   zdialog_run(zd,text_dialog_event,text_dialog_compl);
   
   return;
}

int text_dialog_event(zdialog *zd,const char *event)
{
   GtkWidget   *font_dialog;
   char        *pp, text[1000] = "";
   
   if (strEqu(event,"erase")) {                                            //  erase text entry box
      zdialog_stuff(zd,"text","");
      *pendtext = 0;
      return 0;
   }
   
   if (strEqu(event,"font")) {                                             //  select a new font
      font_dialog = gtk_font_selection_dialog_new(ZTX("select font"));
      gtk_font_selection_dialog_set_font_name(GTK_FONT_SELECTION_DIALOG(font_dialog),textfont);
      gtk_dialog_run(GTK_DIALOG(font_dialog));
      pp = gtk_font_selection_dialog_get_font_name(GTK_FONT_SELECTION_DIALOG(font_dialog));
      if (pp) strncpy0(textfont,pp,59);
      gtk_widget_destroy(GTK_WIDGET(font_dialog));
   }

   zdialog_fetch(zd,"text",text,999);                                      //  capture entered text
   zdialog_fetch(zd,"transp",transp);                                      //  and transparency choice      v.1.6
   zdialog_fetch(zd,"black",txcolor);                                      //  and black/white choice       v.1.6
   if (txcolor != 1) txcolor = 2;

   if (*text) {
      if (NPL == maxNF-1) {
         zmessageACK(ZTX("exceed %d files"),maxNF);
         return 0;
      }
      strncpy0(pendtext,text,999);                                         //  save for layout insertion
      *pendfile = 0;
   }

   return 0;
}

int text_dialog_compl(zdialog *zd, int zstat)
{
   zdialog_free(zd);
   return 0;
}


//  save current image directory upon exit, reload upon startup            //  v.1.2
//  directory is saved in file $HOME/.printoxx/image_directory

void save_imagedirk()
{
   char     command[maxfcc+100];
   int      ignore;
   
   snprintf(command,maxfcc+99,"echo %s > %s/image_directory",              //  save imagedirk to file
                                    imagedirk, get_zuserdir());
   ignore = system(command);
   return;
}

void load_imagedirk()
{
   int            err;
   FILE           *fid;
   struct stat    statdat;
   char           dirbuff[maxfcc], *pp;

   pp = getcwd(imagedirk,maxfcc-1);                                        //  default is current directory

   snprintf(dirbuff,maxfcc-1,"%s/image_directory",get_zuserdir());         //  read saved file
   fid = fopen(dirbuff,"r");
   if (! fid) return;
   pp = fgets_trim(dirbuff,maxfcc-1,fid,1);
   fclose(fid);
   if (! pp) return;
   err = stat(dirbuff,&statdat);                                           //  contains valid directory name?
   if (err) return;
   if (! S_ISDIR(statdat.st_mode)) return;
   strcpy(imagedirk,dirbuff);                                              //  yes, use it
   return;
}


//  make a layout pixbuf containing all scaled images
//  layout 1 is for editing, scaled pixbufs are saved and reused
//  layout 2 is for printing or saving as a file - 2x size of layout1

GdkPixbuf * make_layout1()
{
   GdkPixbuf   *pixbuf1, *pixbuf2;
   int         ii, ww, hh, type, transp, txcolor;
   int         xpos, ypos, orgx, orgy;
   double      rscale;
   
   ww = layout_ww;                                                         //  make layout pixbuf
   hh = layout_hh;
   pixbuf1 = gdk_pixbuf_new(colorspace,0,8,ww,hh);
   gdk_pixbuf_fill(pixbuf1,(unsigned) 0xffffffff);
   
   for (ii = 0; ii < NPL; ii++)                                            //  copy image pixbufs
   {                                                                       //    into layout pixbuf
      type = layout[ii].type;
      transp = layout[ii].transp;
      txcolor = layout[ii].txcolor;
      rscale = layout[ii].rscale;
      ww = int(rscale * layout[ii].ww + 0.5);
      hh = int(rscale * layout[ii].hh + 0.5);
      if (! layout[ii].pixbuf2) {
         pixbuf2 = pixbuf_scale(layout[ii].pixbuf1,ww,hh,bilinear);        //  scale image to layout
         layout[ii].pixbuf2 = pixbuf2;                                     //  retain for efficiency
      }
      pixbuf2 = layout[ii].pixbuf2;    

      xpos = layout[ii].xpos;
      ypos = layout[ii].ypos;

      orgx = orgy = 0;
      if (xpos < 0) {
         orgx = -xpos;
         xpos = 0;
      }
      if (ypos < 0) {
         orgy = -ypos;
         ypos = 0;
      }
      if (xpos + ww > layout_ww) ww = layout_ww - xpos;
      if (ypos + hh > layout_hh) hh = layout_hh - ypos;

      if (type == 1 || ! transp)
         gdk_pixbuf_copy_area(pixbuf2,orgx,orgy,ww-orgx,hh-orgy,pixbuf1,xpos,ypos);
      else  pixbuf_copy_alpha(pixbuf2,orgx,orgy,ww-orgx,hh-orgy,pixbuf1,xpos,ypos,txcolor);
   }
   
   return pixbuf1;
}

GdkPixbuf * make_layout2()                                                 //  make 2x size layout
{
   GdkPixbuf   *pixbuf1, *pixbuf2;
   int         ii, ww, hh, type, transp, txcolor;
   int         xpos, ypos, orgx, orgy;
   double      rscale;
   
   ww = 2 * layout_ww;
   hh = 2 * layout_hh;
   pixbuf1 = gdk_pixbuf_new(colorspace,0,8,ww,hh);
   gdk_pixbuf_fill(pixbuf1,(unsigned) 0xffffffff);
   
   for (ii = 0; ii < NPL; ii++)                                            //  copy image pixbufs
   {                                                                       //    into layout pixbuf
      type = layout[ii].type;
      transp = layout[ii].transp;
      txcolor = layout[ii].txcolor;
      rscale = 2 * layout[ii].rscale;
      ww = int(rscale * layout[ii].ww + 0.5);
      hh = int(rscale * layout[ii].hh + 0.5);
      pixbuf2 = pixbuf_scale(layout[ii].pixbuf1,ww,hh,bilinear);           //  scale image to layout

      xpos = 2 * layout[ii].xpos;
      ypos = 2 * layout[ii].ypos;

      orgx = orgy = 0;
      if (xpos < 0) {
         orgx = -xpos;
         xpos = 0;
      }
      if (ypos < 0) {
         orgy = -ypos;
         ypos = 0;
      }
      if (xpos + ww > 2 * layout_ww) ww = 2 * layout_ww - xpos;
      if (ypos + hh > 2 * layout_hh) hh = 2 * layout_hh - ypos;

      if (type == 1 || ! transp)
         gdk_pixbuf_copy_area(pixbuf2,orgx,orgy,ww-orgx,hh-orgy,pixbuf1,xpos,ypos);
      else  pixbuf_copy_alpha(pixbuf2,orgx,orgy,ww-orgx,hh-orgy,pixbuf1,xpos,ypos,txcolor);

      g_object_unref(pixbuf2);
   }
   
   return pixbuf1;
}


//  validate an image file and load pixbuf from file                       //  v.1.2
//  if an alpha channel is present, remove it

GdkPixbuf * load_pixbuf(const char *file)
{
   GdkPixbuf   *pxb91 = 0, *pxb92 = 0;
   int         ww91, hh91, rs91, rs92;
   int         nch, nbits, px, py, alfa;
   pixel       ppix91, pix91, ppix92, pix92;

   pxb91 = gdk_pixbuf_new_from_file(file,gerror);                          //  validate file and load pixbuf
   if (! pxb91) return 0;

   nch = gdk_pixbuf_get_n_channels(pxb91);
   nbits = gdk_pixbuf_get_bits_per_sample(pxb91);
   alfa = gdk_pixbuf_get_has_alpha(pxb91);

   if (nch < 3 || nbits != 8) {                                            //  must be 3 or 4 channels
      g_object_unref(pxb91);                                               //  and 8 bits per channel
      return 0;
   }
   
   if (! alfa) return pxb91;                                               //  no alpha channel, 3 channels

   ww91 = gdk_pixbuf_get_width(pxb91);                                     //  copy without alpha    v.1.2
   hh91 = gdk_pixbuf_get_height(pxb91);
   rs91 = gdk_pixbuf_get_rowstride(pxb91);
   ppix91 = gdk_pixbuf_get_pixels(pxb91);

   pxb92 = gdk_pixbuf_new(colorspace,0,8,ww91,hh91);
   rs92 = gdk_pixbuf_get_rowstride(pxb92);
   ppix92 = gdk_pixbuf_get_pixels(pxb92);

   for (py = 0; py < hh91; py++)
   for (px = 0; px < ww91; px++)
   {
      pix91 = ppix91 + rs91 * py + 4 * px;
      pix92 = ppix92 + rs92 * py + 3 * px;
      pix92[0] = pix91[0];
      pix92[1] = pix91[1];
      pix92[2] = pix91[2];
   }

   g_object_unref(pxb91);
   return pxb92;
}


//  create a pixbuf containing text
//  drawWin is from gtk_drawing_area_new()
//  drawWin must be a realized window even though it is not modified

GdkPixbuf * pixbuf_text(GtkWidget *drawWin, cchar *text, cchar *font, int color)
{
   PangoFontDescription  *pfont;
   static GdkColormap    *colormap = 0;
   static GdkColor       black, white;
   GdkColor              foreground, background;

   PangoLayout    *playout;
   GdkPixmap      *pixmap;
   GdkGC          *gdkgc;
   GdkPixbuf      *pixbuf;
   int            ww, hh;

   if (! colormap) {
      black.red = black.green = black.blue = 0;
      white.red = white.green = white.blue = 0xffff;
      colormap = gtk_widget_get_colormap(drawWin);
      gdk_rgb_find_color(colormap,&black);
      gdk_rgb_find_color(colormap,&white);
   }
   
   if (color == 1) {
      foreground = black;
      background = white;
   }
   else {
      foreground = white;
      background = black;
   }

   pfont = pango_font_description_from_string(font);
   playout = gtk_widget_create_pango_layout(drawWin,null);
   pango_layout_set_font_description(playout,pfont);
   pango_layout_set_text(playout,text,-1);
   pango_layout_get_pixel_size(playout,&ww,&hh);
   if (! ww) return 0;

   pixmap = gdk_pixmap_new(drawWin->window,ww,hh,-1);
   if (! pixmap) zappcrash("cannot create pixmap");
   gdkgc = gdk_gc_new(pixmap);
   gdk_gc_set_foreground(gdkgc,&background);
   gdk_draw_rectangle(pixmap,gdkgc,1,0,0,ww,hh);

   gdk_draw_layout_with_colors(pixmap,gdkgc,0,0,playout,&foreground,&background);
   pixbuf = gdk_pixbuf_get_from_drawable(null,pixmap,0,0,0,0,0,ww,hh);
   if (! pixbuf) zappcrash("cannot create text pixbuf");
   
   g_object_unref(playout);
   g_object_unref(pixmap);
   g_object_unref(gdkgc);

   return pixbuf;
}


//  copy pixbuf1 into pixbuf2, treating "color" in pixbuf1 as transparent

void pixbuf_copy_alpha(GdkPixbuf *pxb1, int x1, int y1, int ww, int hh,    //  new function   v.1.6
                       GdkPixbuf *pxb2, int x2, int y2, int color)
{
   int         px, py, px1, py1, px2, py2;
   int         ww1, hh1, rs1, ww2, hh2, rs2;
   pixel       ppix1, ppix2, pix1, pix2;
   int         red1, green1, blue1, red2, green2, blue2;
   double      f1, f2;
   
   if (color == 1) color = 255;                                            //  black text, white is transparent
   else color = 0;                                                         //  white text, black is transparent

   ww1 = gdk_pixbuf_get_width(pxb1);
   hh1 = gdk_pixbuf_get_height(pxb1);
   rs1 = gdk_pixbuf_get_rowstride(pxb1);
   ppix1 = gdk_pixbuf_get_pixels(pxb1);

   ww2 = gdk_pixbuf_get_width(pxb2);
   hh2 = gdk_pixbuf_get_height(pxb2);
   rs2 = gdk_pixbuf_get_rowstride(pxb2);
   ppix2 = gdk_pixbuf_get_pixels(pxb2);
   
   for (px = 0; px < ww; px++)
   for (py = 0; py < hh; py++)
   {
      px1 = x1 + px;                                                       //  copy-from pixel
      py1 = y1 + py;
      pix1 = ppix1 + rs1 * py1 + 3 * px1;

      red1 = pix1[0];
      green1 = pix1[1];
      blue1 = pix1[2];

      if (red1 == color && green1 == color && blue1 == color)              //  transparent, do not copy
         continue;

      px2 = x2 + px;                                                       //  copy-to pixel
      py2 = y2 + py;
      pix2 = ppix2 + rs2 * py2 + 3 * px2;

      red2 = pix2[0];
      green2 = pix2[1];
      blue2 = pix2[2];
      
      f1 = abs(color - red1) / 255.0;                                      //  copy-to pixel mixture
      f2 = 1.0 - f1;
      red2 = int(f1 * red1 + f2 * red2);

      f1 = abs(color - green1) / 255.0;
      f2 = 1.0 - f1;
      green2 = int(f1 * green1 + f2 * green2);

      f1 = abs(color - blue1) / 255.0;
      f2 = 1.0 - f1;
      blue2 = int(f1 * blue1 + f2 * blue2);
      
      pix2[0] = red2;
      pix2[1] = green2;
      pix2[2] = blue2;
   }

   return;
}


